import logging
import sys

import colorama
from diagtest.exceptions import InterpreterError

class LogFormatter(logging.Formatter):
    """Custom log formatter"""

    __slots__ = ('formatters',)

    colors = {
        logging.DEBUG:    "\033[36m",
        logging.INFO:     "\033[32m",
        logging.WARNING:  "\033[33m",
        logging.ERROR:    "\033[31m",
        logging.CRITICAL: "\033[1m\033[31m",
    }

    time_fmt = "%Y-%m-%d %H:%M:%S"

    @staticmethod
    def get_format(color: str) -> str:
        """Returns a format string with the appropriate color.

        Args:
            color (str): ANSI escape code

        Returns:
            str: fmt string
        """

        reset = "\033[0m"
        return f"(%(asctime)s) {color}%(brace_l)s%(filename)s:%(lineno)d%(brace_r)s: %(message)s{reset}"

    def __init__(self):
        super().__init__()

        self.formatters = {
            level: logging.Formatter(self.get_format(color), self.time_fmt)
            for level, color in LogFormatter.colors.items()
        }

    def format(self, record: logging.LogRecord) -> str:
        builtin = record.name != 'root'
        record.brace_l = '[' if builtin else '('
        record.brace_r = ']' if builtin else ')'
        formatter = self.formatters.get(record.levelno)
        assert formatter
        return formatter.format(record)


def setup_logger() -> None:
    """Enables this custom logger globally."""
    colorama.init()

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(LogFormatter())

    logging.basicConfig(level=logging.INFO, handlers=[handler], force=True)
    logging.root.addHandler(handler)

    sys.excepthook = handle_exception

def handle_exception(type_, exception, trace):
    """Global exception handler.
    Prints uncaught exceptions with the custom formatter.

    Args:
        type_: Exception type
        exception: Exception value
        trace: Exception trace
    """
    logger = logging.getLogger(__package__)

    if type_ is InterpreterError or issubclass(type_, InterpreterError):
        logger.error("Traceback:")
        logger.error('  File: "%s", line %d', exception.file, exception.line)
        logger.error("%s: %s", type_.__name__, str(exception))

    if logger.isEnabledFor(logging.DEBUG):
        logger.exception("Exception occurred: ", exc_info=(type_, exception, trace))
    else:
        logger.error("Exception occurred: %s %s", type_.__name__, exception)

    raise SystemExit(1)



#!/usr/bin/env python3
''' Allow running diagtest as module: `python -m diagtest` '''

from diagtest import main

if __name__ == "__main__":
    # click edits the parameters for this call, however pylint cannot possibly know about this
    main() # pylint: disable=no-value-for-parameter



import re

from dataclasses import dataclass
from typing import Protocol, Optional
from diagtest.report import Level
from diagtest.util import Result

from diagtest.compiler import ProcessedResult


class Assertion(Protocol):
    def check(self, result: Result) -> bool:
        ...


class Message:
    def __init__(self, level: Level, text: Optional[str] = None, regex: Optional[str | re.Pattern] = None):
        self.regex = regex is not None
        self.text = re.compile(regex) if regex is not None else text
        self.level = level

    def check(self, result: ProcessedResult):
        def check_one(message):
            if self.regex:
                assert self.text is not None
                return re.match(self.text, message) is not None
            return self.text == message

        return any(check_one(diagnostic.message) for diagnostic in result.diagnostics[self.level])

    def __repr__(self):
        require = f"REQUIRE {self.level.name}"
        if self.regex:
            return f"{require} MATCHES {self.text.pattern if isinstance(self.text, re.Pattern) else str(self.text)}"
        return f"{require}: {str(self.text)}"

@dataclass(frozen=True)
class ReturnCode:
    expected: int

    def check(self, result: Result):
        return result.returncode == self.expected

    def __repr__(self):
        return f"RETURNS {self.expected}"

@dataclass(frozen=True)
class ErrorCode:
    expected: str

    def check(self, result: Result):
        return any(diagnostic.error_code is not None and diagnostic.error_code == self.expected
                    for diagnostics in result.diagnostics.values()
                    for diagnostic in diagnostics)

    def __repr__(self):
        return f"RAISES {self.expected}"



import logging
import re
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Any

import em

from diagtest.assertion import Message, ReturnCode, ErrorCode
from diagtest.test import Test, TestSuiteResult
from diagtest.compiler import Compiler
from diagtest.report import Level
from diagtest.exceptions import UsageError
from diagtest.util import change_repr
from diagtest.default import compilers, languages
from diagtest.language import Language, detect_language

@dataclass
class TestSuite:
    tests: list[Test]
    source: Path

    def run(self):
        return TestSuiteResult([test.run(self.source) for test in self.tests], source_file=self.source)

class Runner:
    def __init__(self, sources: Path | list[Path], out_path: Optional[Path] = None, language: str = ""):
        self.out_path = out_path
        self.language = language
        self.sources = sources if isinstance(sources, list) else [sources]

    def expand(self, source: Path):
        out_path = self.out_path or source.parent / "build"

        out_path.mkdir(exist_ok=True, parents=True)
        with Parser(source, self.language) as (processed, tests):
            preprocessed_source = out_path / source.name
            preprocessed_source.write_text(processed)
            return TestSuite(tests, preprocessed_source)

    def run(self):
        for source in self.sources:
            try:
                suite = self.expand(source)
                yield suite.run()
            except Exception as exc: # TODO refine
                logging.error("Running %s failed. Reason: %s", source, str(exc))


class Parser:
    def __init__(self, source: Path, language: str = ""):
        self.globals = {name: getattr(self, name)
                        for name in dir(type(self))
                        if not name.startswith('_')}

        self.interpreter = em.Interpreter(globals=self.globals)

        self.tests: list[Test] = []
        self.include_paths: list[Path] = []
        self.source: Path = source
        self.language: str = language
        self.language_definition: Optional[type[Language]] = None

        if not self.language:
            candidates = detect_language(source)
            if len(candidates) > 1:
                logging.debug("Language for %s is ambiguous. Candidates: %s", source, candidates)
            elif len(candidates) == 1:
                self.language = candidates[0]
            else:
                logging.debug("Could not detect language for %s", source)

        self.set_language(language)


    def __enter__(self):
        content = self.source.read_text(encoding="utf-8")
        processed = self.interpreter.expand(content, name=self.source)
        return processed, self.tests

    def __exit__(self, type_, value, traceback):
        self.interpreter.shutdown()

    def _resolve_path(self, path: Path | str):
        if not isinstance(path, Path):
            path = Path(path)

        if not path.is_absolute():
            file, *_ = self.interpreter.identify()
            path = Path(file).parent / path
        return path.resolve()

    def include(self, path: Path | str):
        self.interpreter.include(str(self._resolve_path(path)))

    def include_path(self, path: Path | str):
        """
            Adds a directory to the include path. This may only be called before load_defaults!
        Args:
            path: Directory to add to the include path
        """
        path = self._resolve_path(path)
        assert path.exists(), "Include path does not exist"
        assert path.is_dir(), "Path does not point to a directory"
        self.include_paths.append(path)

    def set_language(self, language: str = ""):
        if not language:
            return

        self.language = language
        definition = languages.get(language)
        if self.language_definition != definition:
            self.language_definition = definition
            # TODO reload interpreter

    def load_defaults(self, language: str = "", dialect: str = ""):
        if language:
            self.set_language(language)
        assert self.language, "Must specify language to load defaults for"
        defaults: dict[str, Any] = {}

        def wrap(cls):
            def inner(**kwargs):
                if 'language' not in kwargs:
                    kwargs['language'] = self.language

                if self.include_paths:
                    if 'args' not in kwargs:
                        kwargs['args'] = []

                    assert hasattr(cls, "include"), f"Compiler {cls} cannot expand includes."
                    for path in self.include_paths:
                        kwargs['args'].append(cls.include(path))

                nonlocal dialect
                if dialect and 'dialect' not in kwargs:
                    kwargs['dialect'] = dialect

                return cls(**kwargs)
            return inner

        for compiler in compilers:
            if self.language not in getattr(compiler, 'languages', []):
                continue

            defaults[compiler.__name__] = wrap(compiler)
            defaults[compiler.__name__.lower()] = defaults[compiler.__name__]()

        if not defaults:
            logging.warning("Could not find defaults for language %s", self.language)
            return
        self.update_globals(defaults)

    def update_globals(self, new_globals: dict[str, Any]):
        self.interpreter.updateGlobals(new_globals)

    def test(self, name: str):
        assert self.language, "Automatic language detection failed. "\
            "Specify it as command line argument or use set_language before the first test"
        assert self.language_definition is not None, "Missing language definition"
        self.tests.append(Test(self.language_definition.identifier(name), name))

        def report_usage_error():
            raise UsageError(self.interpreter.identify(),
                             "Make sure to NOT place a space before the curly brace after @test(...)")

        @change_repr(report_usage_error)
        def wrap(code: str):
            nonlocal name
            assert self.language_definition is not None, "Missing language definition"
            return self.language_definition.wrap_test(name, code)

        return wrap

    def return_code(self, compiler: Compiler, code: int):
        self.tests[-1].add_assertion(compiler, ReturnCode(code))

    def error_code(self, compiler: Compiler, code: str):
        self.tests[-1].add_assertion(compiler, ErrorCode(code))

    def note(
        self,
        compiler: Compiler,
        text: Optional[str] = None,
        *,
        regex: re.Pattern[str] | str | None = None,
    ):
        self.tests[-1].add_assertion(compiler, Message(Level.note, text, regex))

    def warning(
        self,
        compiler: Compiler,
        text: Optional[str] = None,
        *,
        regex: re.Pattern[str] | str | None = None,
    ):
        self.tests[-1].add_assertion(compiler, Message(Level.warning, text, regex))

    def error(
        self,
        compiler: Compiler,
        text: Optional[str] = None,
        *,
        regex: re.Pattern[str] | str | None = None,
    ):
        self.tests[-1].add_assertion(compiler, Message(Level.error, text, regex))

    def fatal_error(
        self,
        compiler: Compiler,
        text: Optional[str] = None,
        *,
        regex: re.Pattern[str] | str | None = None,
    ):
        self.tests[-1].add_assertion(compiler, Message(Level.fatal_error, text, regex))



from diagtest.language import Language

class C(Language):
    languages = 'c', 'gnu'
    suffixes = ['.c', '.h', '.i']
    @classmethod
    def identifier(cls, name: str):
        return f"TEST_{''.join(c if c.isalnum() else '_' for c in name.upper())}"

    @classmethod
    def wrap_test(cls, name: str, code: str):
        return f"\n#ifdef {cls.identifier(name)}\n{code}\n#endif"

class CPP(C):
    languages = 'c++', 'gnu++'
    suffixes = [*C.suffixes, '.C', '.cc', '.cpp', '.CPP', '.c++', '.cp', '.cxx', '.hh', '.hpp', '.H', '.tcc', '.ii']



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional


class Level(Enum):
    # TODO this should be an open set instead of an enum
    note = "note"
    warning = "warning"
    error = "error"
    fatal_error = "fatal error"
    # ice = auto() # TODO


@dataclass
class SourceLocation:
    path: Path
    line: Optional[int] = None
    column: Optional[int] = None


@dataclass
class Diagnostic:
    message: str
    source_location: Optional[SourceLocation] = None
    error_code: Optional[str] = None  # MSVC specific




import re
from pathlib import Path
from functools import cache

from diagtest.compilers.gcc import GCC
from diagtest.dialect import Dialect
from diagtest.util import run

class Clang(GCC):
    languages = 'c', 'c++', 'gnu', 'gnu++'
    executable_pattern = r"^clang(-[0-9]+)?(\.exe|\.EXE)?$"

    standard_pattern = re.compile(r"use '(?P<standard>[^']+)'")
    standard_alias_pattern = re.compile(r"(( or|,) '(?P<alias>[^']+))")

    @staticmethod
    @cache
    def _get_supported_dialects(compiler, language: str):
        result = run([str(compiler), f'-x{language}', '-std=dummy', '-'])
        for line in result.stderr.splitlines():
            standard_match = re.search(Clang.standard_pattern, line)
            if standard_match is None:
                continue
            standard = standard_match['standard']
            aliases = [match['alias'] for match in Clang.standard_alias_pattern.finditer(line)]
            yield Dialect(standard, aliases)

    @staticmethod
    def filter_gnu(compiler, language: str):
        gnu_name = f'gnu{language[1:]}'
        result: dict[str, list[tuple[str, ...]]] = {gnu_name: [], language: []}
        for standard in Clang._get_supported_dialects(compiler, language):
            is_gnu = any(name.startswith('gnu') for name in standard)
            result[gnu_name if is_gnu else language].append(standard)
        return result

    @staticmethod
    @cache
    def get_supported_dialects(compiler: Path):
        return {**Clang.filter_gnu(compiler, 'c'),
                **Clang.filter_gnu(compiler, 'c++')}



from collections import defaultdict
import re

from abc import ABC, abstractmethod
from functools import cache
from operator import attrgetter
from pathlib import Path
from typing import Any, Iterable, Optional, Type

from packaging.specifiers import SpecifierSet
from packaging.version import Version

from diagtest.report import Diagnostic, Level, SourceLocation
from diagtest.util import Result, which, remove_duplicates, run


class CompilerInfo:
    __slots__ = "executable", "languages", "version", "target"

    def __init__(self, executable: Path, languages: list[str], version: str | Version, target: Optional[str] = None):
        self.executable: Path = executable
        self.languages: list[str] = languages
        self.version = version if isinstance(version, Version) else Version(version)
        self.target: Optional[str] = target

    def dump(self) -> dict[str, Any]:
        info = {'executable': self.executable,
                'version': self.version}
        if self.target:
            info['target'] = self.target

        info['languages'] = self.languages
        return info

    def __str__(self):
        return str(self.dump())

    @property
    def version_string(self):
        ret = str(self.version)
        if self.target:
            ret += f" {self.target}"
        return ret

    def has_executable(self, executable: Path):
        return self.executable == executable

    def has_language(self, language: str):
        return language in self.languages

    def has_version(self, specifier: SpecifierSet):
        return self.version in specifier

    def has_target(self, query: str | re.Pattern):
        if self.target is None:
            return False

        if isinstance(query, str):
            return query == self.target
        elif isinstance(query, re.Pattern):
            return re.match(query, self.target)
        raise RuntimeError("Query must be string or regex pattern")


class CompilerCollection(list[CompilerInfo]):
    def __getattr__(self, attr: str):
        """ Filter by arbitrary attribute.
        This assumes CompilerInfo has been subclassed and provides a method prefixed with has_
        to determine if any compilers match the search query
        """
        if not attr.startswith('by_'):
            # Ignore other unknown attributes
            return None

        def apply_filter(query: Any):
            nonlocal attr
            filtered_compilers: list[CompilerInfo] = []
            for compiler in self:
                check = getattr(compiler, attr.replace('by_', 'has_'), None)
                if not check:
                    continue

                if check(query):
                    filtered_compilers.append(compiler)

            return CompilerCollection(filtered_compilers)

        return apply_filter

    @property
    def executables(self):
        return [compiler.executable for compiler in self]


class ProcessedResult(Result):
    def __init__(self, result: Result, name: str, compiler: CompilerInfo):
        super().__init__(result.command, result.returncode, result.stdout, result.stderr, result.start_time, result.end_time)
        self.name = name
        self.compiler = compiler
        # TODO Level should be an open set. Replace with str
        self.diagnostics: dict[Level, list[Diagnostic]] = defaultdict(list)

    def extend(self, diagnostics: Iterable[tuple[Level, Diagnostic]]):
        for level, diagnostic in diagnostics:
            self.diagnostics[level].append(diagnostic)


class Compiler(ABC):
    def __init__(
        self,
        /, *,
        language: str = "",
        executable: Optional[Path] = None,
        version: Optional[SpecifierSet | str] = None,
        target: Optional[str | re.Pattern] = None,
        args: Optional[list[str]] = None
    ):
        self.options = args or []

        assert hasattr(type(self), "languages"), "Compiler definition lacks languages attribute"
        if language not in getattr(self, "languages", []):
            raise RuntimeError(f"Cannot run compiler {self!s} in language mode {language}")
        self.language = language

        assert hasattr(self, "diagnostic_pattern"), "Compiler definition lacks a diagnostic pattern"
        self.diagnostic_pattern = getattr(self, 'diagnostic_pattern')  # hack to make mypy happy
        if isinstance(self.diagnostic_pattern, str):
            # compile the pattern if it hasn't yet happened
            self.diagnostic_pattern = re.compile(self.diagnostic_pattern)

        if executable:
            self.compilers = CompilerCollection([self.get_info(executable)])
        else:
            discovered_paths = self.discover()
            compiler_paths = discovered_paths[self.language] if isinstance(discovered_paths, dict) else discovered_paths
            self.compilers = CompilerCollection(self.get_info(path) for path in compiler_paths)

        if version is not None:
            version = SpecifierSet(version) if isinstance(version, str) else version
            self.compilers = self.compilers.by_version(version)

        if target is not None:
            self.compilers = self.compilers.by_target(target)

    def __call__(self, language: Optional[str] = None, args: Optional[list[str]] = None, **kwargs):
        return type(self)(language=language or self.language, args=args or self.options, **kwargs)

    @classmethod
    def __init_subclass__(cls, **kwargs):
        if cls not in compilers and hasattr(cls, 'languages'):
            compilers.append(cls)
        super().__init_subclass__(**kwargs)

    def extract_diagnostics(self, lines):
        for line in lines.splitlines():
            diagnostic = re.match(self.diagnostic_pattern, line)
            if not diagnostic:
                continue

            parts = diagnostic.groupdict()
            source_location = (
                SourceLocation(
                    parts["path"],
                    parts.get("line", None),
                    parts.get("column", None),
                )
                if "path" in parts
                else None
            )
            level = Level(parts["level"])
            yield level, Diagnostic(parts["message"], source_location, parts.get("error_code", None))

    def get_name(self, info: CompilerInfo):
        name = type(self).__name__
        target_str = f"{info.target} " if info.target else ""
        return f"{target_str}{name} ({info.version})"

    def execute(self, name: str, compiler: CompilerInfo, source_file: Path,
                extra_args: Optional[list[str]] = None, env: Optional[dict[str, str]] = None) -> ProcessedResult:

        result = run([str(compiler.executable.resolve()),
                      # TODO select language if needed
                      *self.get_compile_options(),
                      *(extra_args or []),
                      str(source_file.resolve())], env=env)

        processed = ProcessedResult(result, name, compiler)
        processed.extend(self.extract_diagnostics(result.stderr))
        processed.extend(self.extract_diagnostics(result.stdout))
        return processed

    def run_test(self, source_file: Path, test: str):
        for compiler in self.compilers:
            yield self.execute(self.get_name(compiler), compiler, source_file, [self.select_test(test)])

    @classmethod
    @cache
    def discover(cls) -> list[Path] | dict[str | tuple[str, ...], list[Path]]:
        assert hasattr(cls, 'executable_pattern'), "Auto discovery failed. Compiler definition lacks executable_pattern"
        executable_pattern = getattr(cls, 'executable_pattern')
        if isinstance(executable_pattern, dict):
            # alternatives for various languages
            return {language: remove_duplicates(which(pattern))
                    for language, pattern in executable_pattern.items()}
        return remove_duplicates(which(executable_pattern))

    @staticmethod
    @abstractmethod
    def select_language(language: str):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def select_test(test: str):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_version(path: Path) -> Version | str:
        raise NotImplementedError()

    @staticmethod
    def get_target(path: Path) -> Optional[str]:
        return None

    @classmethod
    def get_info(cls, path: Path) -> CompilerInfo:
        assert hasattr(cls, "languages"), "Compiler definition lacks languages attribute"
        return CompilerInfo(executable=path,
                            languages=getattr(cls, "languages"),
                            version=cls.get_version(path),
                            target=cls.get_target(path))

    def get_compile_options(self):
        options = self.options or []
        if language := self.select_language(self.language):
            options.append(language)
        return options

    def get_alternative(self, path: Path) -> Path:
        return path

    def __str__(self):
        name = type(self).__name__
        versions = ", ".join(str(compiler.version) for compiler in sorted(self.compilers, key=attrgetter('version')))
        return f"{name} ({versions})" if versions else name

    def __or__(self, other):
        if isinstance(other, (list, Collection)):
            other.append(self)
            return other
        return Collection([self, other])

    @property
    def available(self):
        return len(self.compilers) != 0


class Collection(list[Compiler]):
    def __or__(self, other):
        if isinstance(other, (list, Collection)):
            self.extend(other)
        else:
            self.append(other)
        return self

    def run_test(self, file: Path, test_id: str):
        for compiler in self:
            yield compiler.run_test(file, test_id)

    @property
    def available(self):
        return len(self) != 0

    def __hash__(self):
        return hash(iter(self))

    def __str__(self):
        if len(self) == 1:
            return str(self[0])

        return f"[{', '.join(str(data) for data in self)}]"


compilers: list[Type[Compiler]] = []



from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from diagtest.assertion import Assertion
from diagtest.compiler import Compiler, ProcessedResult


@dataclass
class RunResult:
    compiler: Compiler
    result: ProcessedResult | None
    assertions: dict[Assertion, bool | None]

Result = tuple[ProcessedResult | None, bool | None]

@dataclass
class TestResult:
    name: str
    runs: list[RunResult]

    def by_compiler(self):
        ret: dict[Compiler, dict[Assertion, list[Result]]] = defaultdict(lambda: defaultdict(list))
        for run in self.runs:
            for assertion, success in run.assertions.items():
                ret[run.compiler][assertion].append((run.result, success))
        return dict(ret)

    def by_assertion(self):
        ret: dict[Assertion, dict[Compiler, list[Result]]] = defaultdict(lambda: defaultdict(list))
        for run in self.runs:
            for assertion, success in run.assertions.items():
                ret[assertion][run.compiler].append((run.result, success))
        return dict(ret)

    def __bool__(self):
        return all(success if success is not None else True # count skipped as successful
                   for run in self.runs
                   for success in run.assertions.values())

@dataclass
class Test:
    identifier: str
    name: str
    assertions: defaultdict[Compiler, list[Assertion]] = field(default_factory=lambda: defaultdict(list))

    @property
    def compilers(self):
        return self.assertions.keys()

    def add_assertion(self, compiler: Compiler, assertion: Assertion):
        self.assertions[compiler].append(assertion)

    def run(self, source: Path):
        results = {compiler: list(compiler.run_test(source, self.identifier)) for compiler in self.compilers}

        def run_assertions():
            for compiler, assertions in self.assertions.items():
                if not compiler.available:
                    yield RunResult(compiler, None, {assertion: None for assertion in assertions})
                    continue

                for result in results[compiler]:
                    yield RunResult(compiler, result, {assertion: assertion.check(result) for assertion in assertions})

        return TestResult(self.name, runs=list(run_assertions()))

class TestSuiteResult(list[TestResult]):
    __slots__ = 'source_file'

    def __init__(self, *args, source_file: Optional[Path] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.source_file = source_file



import logging
from abc import abstractmethod
from pathlib import Path
from typing import Type

from diagtest.util import UniqueDict


class Language:
    languages: tuple[str, ...] | list[str]
    suffixes: tuple[str, ...] | list[str]

    @classmethod
    def __init_subclass__(cls):
        if not hasattr(cls, 'languages'):
            logging.warning("Language definition {cls.__name__} lacks the `languages` attribute.")
            return

        for language in getattr(cls, 'languages', []):
            languages[language] = cls

    @classmethod
    @abstractmethod
    def identifier(cls, name: str) -> str:
        ...

    @classmethod
    @abstractmethod
    def wrap_test(cls, name: str, code: str) -> str:
        ...


languages: dict[str, Type[Language]] = UniqueDict()


def detect_language(source_file: Path):
    return [language
            for language, definition in languages.items()
            if source_file.suffix in definition.suffixes]



import logging
from pathlib import Path
from itertools import groupby
from functools import partial

from diagtest.compiler import compilers
from diagtest.language import languages
from diagtest.util import remove_duplicates, print_dict
__all__ = ['compilers', 'languages', 'load_compilers', 'load_languages']

def dump_compilers():
    def key_function(compiler, path: Path):
        return compiler.get_version(path), compiler.get_target(path)

    for compiler in compilers:
        print(f"Compiler `{compiler.__name__}`")
        discovered = compiler.discover()
        if not discovered:
            # print("    not found")
            continue
        if isinstance(discovered, dict):
            discovered = remove_duplicates([compiler
                                            for compiler_list in discovered.values()
                                            for compiler in compiler_list])

            for _, group in groupby(discovered, partial(key_function, compiler)):
                executables = list(group)
                assert executables, "Empty group in compilers listed by version. How?"
                info = compiler.get_info(executables[0]).dump()
                info['executable'] = executables
                print_dict(info, indent_amount=4, indent_level=1)
            continue

        for compiler_path in discovered:
            info = compiler.get_info(compiler_path).dump()
            print_dict(info, indent_amount=4, indent_level=1)


def load_compilers():
    # TODO load all files in compilers directory instead
    import diagtest.compilers.gcc
    import diagtest.compilers.msvc
    import diagtest.compilers.clang


def load_languages():
    import diagtest.languages.cfamily



import re
from pathlib import Path
from functools import cache
from collections import defaultdict
from contextlib import suppress

from diagtest.dialect import DialectCompiler, Dialect
from diagtest.util import run


class GCC(DialectCompiler):
    languages = 'c', 'c++', 'gnu', 'gnu++'
    diagnostic_pattern = r"^((?P<path>.*?):((?P<line>[0-9]+):)?((?P<column>[0-9]+):)? )?"\
                         r"((?P<level>error|warning|note): )(?P<message>.*)$"

    version_pattern = re.compile(r"((Target: (?P<target>.*))|(Thread model: (?P<thread_model>.*))|"
                                 r"((gcc|clang) version (?P<version>[0-9\.]+)))")

    gcc_pattern = r"^gcc(-[0-9]+)?(\.exe|\.EXE)?$"
    gcc_cpp_pattern = r"^g\+\+(-[0-9]+)?(\.exe|\.EXE)?$"

    executable_pattern: str | dict[str, str] = {
        'c': gcc_pattern,
        'gnu': gcc_pattern,
        'c++': gcc_cpp_pattern,
        'gnu++': gcc_cpp_pattern
        #'fortran': r"^gfortran(-[0-9]+)?(\.exe|\.EXE)?$" # TODO
    }

    @staticmethod
    def select_language(language: str):
        return None

    @staticmethod
    def select_dialect(dialect: str):
        return f"-std={dialect}"

    @staticmethod
    def select_test(test: str):
        return f"-D{test}"

    @staticmethod
    def include(directory: Path):
        return f"-I{directory.absolute()!s}"

    @staticmethod
    @cache
    def _query_version(path: Path) -> dict[str, str]:
        # invoke gcc -v --version
        result = run([str(path), "-v", "--version"])
        version: dict[str, str] = {}
        for source in result.stderr, result.stdout:
            for match in re.finditer(GCC.version_pattern, source):
                version |= {k: v for k, v in match.groupdict().items() if v}
        return version

    @staticmethod
    @cache
    def _get_supported_dialects(path: Path):
        search_pattern = re.compile(r"-std=(?P<standard>[^\s]+)[\s]*(Conform.*((C|C\+\+)( draft)? standard))"
                                    r"((.|(\n    )\s+)*Same.as(.|(\n    )\s+)*-std=(?P<alias>[^\. ]+))?")
        standards = defaultdict(list)
        # invoke gcc -v --help
        result = run([str(path), "-v", "--help"])
        for match in search_pattern.finditer(result.stdout):
            standard = match['standard']
            if alias := match['alias']:
                standards[alias].append(standard)
            else:
                standards[standard].append(standard)
        return [Dialect(standard, [alias for alias in aliases if alias != standard])
                for standard, aliases in standards.items()]

    @classmethod
    @cache
    def get_supported_dialects(cls, path: Path):
        result: dict[str, list[Dialect]] = defaultdict(list)
        for standard in cls._get_supported_dialects(path):
            is_gnu = any(name.startswith('gnu') for name in standard)
            is_cpp = any('++' in name for name in standard)
            suffix = '++' if is_cpp else ''
            result[f'gnu{suffix}' if is_gnu else f'c{suffix}'].append(standard)

        # GCC developers in their infinite wisdom decided to not list C standards in order
        # at least it is in order for both centuries, so do a little swapping here
        for language in 'c', 'gnu':
            standards = result[language]
            if not standards:
                continue

            idx = next(idx for idx, standard in enumerate(standards) if '9' in standard.name)
            last_century = standards[idx:]

            if language == 'c':
                with suppress(StopIteration):
                    # iso9899:199409 is discovered after c99, but 1994 was before 1999
                    index_iso94 = next(idx for idx, standard in enumerate(last_century) if 'iso9899:199409' in standard)
                    index_c99 = next(idx for idx, standard in enumerate(last_century) if 'c99' in standard)
                    assert index_iso94 > index_c99, "Standards discovered in order. Please open an issue!"
                    last_century[index_c99], last_century[index_iso94] = last_century[index_iso94], last_century[index_c99]
            result[language] = [*last_century, *standards[:idx]]

        return result

    @classmethod
    def get_version(cls, path: Path) -> str:
        info = cls._query_version(path)
        assert 'version' in info, "Automatic version detection failed"
        return info['version']

    @classmethod
    def get_target(cls, path: Path) -> str | None:
        info = cls._query_version(path)
        return info.get('target')

    alternative_needle = 'gcc'
    alternative_replacement = {'c': 'gcc', 'gnu': 'gcc', 'c++': 'g++', 'gnu++': 'g++'}

    def get_alternative(self, path: Path):
        if self.language not in self.alternative_replacement:
            return path

        replacement = path.with_name(path.name.replace(self.alternative_needle,
                                                       self.alternative_replacement[self.language]))
        return replacement if replacement.exists() else path

    def has_alternative(self) -> bool:
        return self.language not in self.alternative_replacement



import logging
import os
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path
import time
from typing import Any, Iterable, Optional, TypeVar


@dataclass
class Result:
    command: str
    returncode: int
    stdout: str
    stderr: str
    start_time: int
    end_time: int

    @property
    def elapsed(self):
        assert self.end_time >= self.start_time, "Invalid time measurements"
        return self.end_time - self.start_time

    @property
    def elapsed_ms(self):
        return self.elapsed / 1e6

    @property
    def elapsed_s(self):
        return self.elapsed / 1e9


def run(command: list[str] | str, env: Optional[dict[str, str]] = None):
    command_str = command if isinstance(command, str) else ' '.join(command)
    logging.debug(command_str)
    start_time = time.monotonic_ns()
    result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        check=False,
        env=env
    )
    end_time = time.monotonic_ns()

    return Result(command_str,
                  result.returncode,
                  result.stdout,
                  result.stderr,
                  start_time,
                  end_time)


def which(query: re.Pattern | str):
    query = re.compile(query) if isinstance(query, str) else query
    env_path = os.environ.get('PATH', os.environ.get('Path', os.defpath))
    paths = [Path(path) for path in env_path.split(os.pathsep)]
    for path in paths:
        if not path.exists():
            continue

        for file in path.iterdir():
            if query.match(file.name):
                # resolving here should get rid of symlink aliases
                yield file.resolve()


def change_repr(repr_fnc):
    class ReprWrapper:
        def __init__(self, fnc):
            self.fnc = fnc

        def __call__(self, *args, **kwargs):
            return self.fnc(*args, **kwargs)

        def __repr__(self):
            nonlocal repr_fnc
            return repr_fnc()

    return ReprWrapper


K = TypeVar('K')
V = TypeVar('V')


class UniqueDict(dict[K, V]):
    def __setitem__(self, index, value):
        assert index not in self, f"Collection already contains {index}, cannot set to {value}"
        super().__setitem__(index, value)


Element = TypeVar('Element')


def remove_duplicates(data: Iterable[Element]) -> list[Element]:
    return [*{entry: None for entry in data}.keys()]

def format_dict(item: dict,
                indent_amount: int = 4,
                indent_level: int = 0,
                dict_separator: str = ": ",
                list_separator: str = ", ",
                break_list: bool = False,
                right_align: bool = False) -> Iterable[str]:
    max_width = max(len(str(key)) for key in item)

    def align(inner_key: Any, prefix: str = "", suffix: str = ""):
        nonlocal max_width, right_align
        diff = ' ' * (max_width - len(str(inner_key)))
        wrapped = f"{prefix}{inner_key}{suffix}"
        return diff + wrapped if right_align else wrapped + diff

    for key, value in item.items():
        line = f"{align(key, suffix=dict_separator)}"
        if isinstance(value, dict):
            yield from indent(line, indent_level, indent_amount)
            yield from format_dict(value, indent_amount, indent_level + 1, dict_separator, list_separator, break_list, right_align)

        elif isinstance(value, (list, tuple, set)):
            if break_list:
                yield from indent(line, indent_level, indent_amount)
                for element in value:
                    yield from indent(str(element), indent_level + 1, indent_amount)
            else:
                yield from indent(line + list_separator.join(str(element) for element in value), indent_level, indent_amount)

        else:
            yield from indent(line + str(value), indent_level, indent_amount)


def print_dict(item: dict, *args, **kwargs):
    for line in format_dict(item, *args, **kwargs):
        print(line)

def indent(element: Any, indent_level: int = 1, indent_amount: int = 4):
    indentation = ' ' * (indent_amount * indent_level)
    for line in str(element).splitlines():
        yield f"{indentation}{line}"

def print_indent(what: Any, indent_level: int = 0, indent_amount: int = 4, **kwargs):
    for line in indent(what, indent_level, indent_amount):
        print(line, **kwargs)



from dataclasses import dataclass
from typing import Optional, Any
from pathlib import Path
from abc import abstractmethod

from diagtest.compiler import Compiler, CompilerInfo
from diagtest.util import remove_duplicates


DialectQuery = str | int | list[int | str] | tuple[str, str]

@dataclass
class Dialect:
    name: str
    aliases: Optional[list[str]] = None

    def __str__(self):
        name = self.name
        if self.aliases:
            name += f" ({', '.join(self.aliases)})"
        return name

    def __iter__(self):
        yield self.name
        if self.aliases is not None:
            yield from self.aliases

class DialectInfo(CompilerInfo):
    __slots__ = ["dialects"]

    def __init__(self, dialects: dict[str, Dialect], **kwargs):
        # map of language -> list of supported dialects (and their aliases)
        self.dialects: dict[str, Dialect] = dialects
        super().__init__(**kwargs)

    def dump(self) -> dict[str, Any]:
        info = super().dump()
        info['dialects'] = self.dialects
        return info

    def expand_dialect(self, language: str, dialect: str | int):
        if isinstance(dialect, str) and self.has_dialect(dialect):
            return dialect

        expanded = f"{language}{dialect}"
        if self.has_dialect(expanded):
            return expanded

        raise RuntimeError(f"Standard {expanded} not found in available dialects")

    def filter_dialects(self, language: str, query: str, dialects=None):
        if dialects is None:
            dialects = self.dialects[language]

        greater = query[0] == '>'
        include = query[1] == '='
        version = self.expand_dialect(language, query[1 + include:])
        try:
            index = next(idx for idx, dialect in enumerate(dialects) if version in dialect)
            index += include ^ greater
            return dialects[index:] if greater else dialects[:index]
        except StopIteration as e:
            raise RuntimeError(f"Could not find value {version} in {dialects}") from e

    def get_dialects(self, language: str, query: Optional[DialectQuery] = None):
        def flatten(dialects):
            return remove_duplicates([dialect.name for dialect in dialects])

        if query is None:
            # get the primary version aliases
            return flatten(self.dialects[language])

        if isinstance(query, list):
            return remove_duplicates([self.expand_dialect(language, dialect) for dialect in query])

        if isinstance(query, tuple):
            assert len(query) == 2, "Only a 2-tuple to specify range is allowed"
            assert query[0][0] == '>' and query[1][0] == '<', "Specify ranges as ('>minimum', '<maximum')"
            return flatten(self.filter_dialects(language, query[1], self.filter_dialects(language, query[0])))

        if isinstance(query, int):
            return [self.expand_dialect(language, query)]

        if query.startswith(('>', '<')):
            return flatten(self.filter_dialects(language, query))

        query = self.expand_dialect(language, query)
        if any(query in aliases for aliases in self.dialects[language]):
            return [query]
        return []

    def has_dialect(self, query: str):
        return any(query in dialect
        for dialects in self.dialects.values()
        for dialect in dialects)


class DialectCompiler(Compiler):
    def __init__(self, language: str, dialect: Optional[dict[CompilerInfo, list[str]] | DialectQuery] = None, **kwargs):
        super().__init__(language=language, **kwargs)
        if isinstance(dialect, dict):
            self.dialects = dialect
        else:
            self.dialects = {}
            for compiler in self.compilers:
                assert isinstance(compiler, DialectInfo)
                self.dialects[compiler] = compiler.get_dialects(language, dialect)

    def __call__(self, language: Optional[str] = None, dialect: Optional[dict[CompilerInfo, list[str]] | DialectQuery] = None,  # type: ignore[override]
                 options: Optional[list[str]] = None, executable: Optional[Path | str] = None):
        return type(self)(language=language or self.language,
                          dialect=dialect or self.dialects,
                          options=options or self.options,
                          executable=executable or self.compilers)

    def has_dialect(self, path: Path, dialect: str):
        info = self.get_info(path)
        return info.has_dialect(dialect)

    def get_dialects(self, path: Path, query: Optional[DialectQuery] = None):
        info = self.get_info(path)
        return info.get_dialects(self.language, query)

    @staticmethod
    @abstractmethod
    def select_dialect(dialect: str):
        raise NotImplementedError

    def run_test(self, source_file: Path, test: str):
        for compiler, dialects in self.dialects.items():
            for dialect in dialects:
                name = f"{self.get_name(compiler)} ({dialect})"
                yield self.execute(name, compiler, source_file, [self.select_dialect(dialect), self.select_test(test)])

    @staticmethod
    @abstractmethod
    def get_supported_dialects(path: Path):
        raise NotImplementedError()

    @classmethod
    def get_info(cls, path: Path) -> DialectInfo:
        dialects = cls.get_supported_dialects(path)
        supported_languages = remove_duplicates([*getattr(cls, "languages", []), *dialects.keys()])
        return DialectInfo(dialects=dialects,
                           executable=path,
                           languages=supported_languages,
                           version=cls.get_version(path),
                           target=cls.get_target(path))



import shutil
import os
import re
import logging
import platform
from functools import cache
from pathlib import Path
from enum import Enum

from diagtest.dialect import DialectCompiler, Dialect
from diagtest.compiler import CompilerInfo
from diagtest.util import run


class VsArch(Enum):
    x86 = 'x86'
    x64 = 'amd64'


class MSVC(DialectCompiler):
    languages = 'c', 'c++'
    diagnostic_pattern = r"^((?P<path>[a-zA-Z0-9:\/\\\._-]*?)\((?P<line>[0-9]+)\): )"\
                         r"((?P<level>fatal error|error|warning) )((?P<error_code>[A-Z][0-9]+): )(?P<message>.*)$"

    '''def execute(self, file: Path, test_id: str):
        # TODO
        env = self.get_env(VsArch[self.get_version(self.compiler)['target']])
        for standard in self.selected_standards:
            name = f"{str(self)} ({standard})"
            yield name, self.compile(file, [f"/std:{standard}", *(self.options or []), f"/D{test_id}"], env=env)'''

    @staticmethod
    def select_language(language: str):
        return None

    @staticmethod
    def select_dialect(dialect: str):
        return f"/std:{dialect}"

    @staticmethod
    def select_test(test: str):
        return f"/D{test}"

    @staticmethod
    def include(directory: Path):
        return f"/I{directory.absolute()!s}"

    @staticmethod
    @cache
    def get_help_text(compiler: Path):
        return run([str(compiler), '/help'])

    @staticmethod
    @cache
    def _query_version(compiler: Path) -> dict[str, str]:
        help_text = MSVC.get_help_text(compiler)
        version_pattern = re.compile(r"Version (?P<version>[0-9\.]+) for (?P<target>.*)")
        result = re.search(version_pattern, help_text.stderr)
        assert result is not None, "Querying compiler version failed"
        return result.groupdict()

    @classmethod
    def get_version(cls, path: Path) -> str:
        info = cls._query_version(path)
        assert 'version' in info, "Automatic version detection failed"
        return info['version']

    @classmethod
    def get_target(cls, path: Path) -> str | None:
        info = cls._query_version(path)
        return info.get('target')

    @staticmethod
    @cache
    def get_supported_dialects(compiler: Path):
        help_text = MSVC.get_help_text(compiler)
        standard_pattern = r"\/std:<(?P<standards>.*)> (?P<language>[^ ]+)"
        standards: dict[str, list[Dialect]] = {}
        for match in re.finditer(standard_pattern, help_text.stdout):
            language = match['language'].lower()
            standards[language] = [Dialect(standard) for standard in match["standards"].split('|')]
        return standards

    @staticmethod
    @cache
    def vswhere():
        # this is a stable path
        vswhere_exe = Path(os.environ["PROGRAMFILES(X86)"]) / "Microsoft Visual Studio" / "Installer" / "vswhere.exe"
        result = run([str(vswhere_exe)])
        assert result.returncode == 0, "Could not run vswhere"

        def search(fields: list[str]):
            for line in result.stdout.splitlines():
                for field in fields:
                    if not line.startswith(field):
                        continue
                    yield field, line.split(f"{field}: ", maxsplit=1)[1]

        return dict(search(['installationPath', 'installationVersion', 'displayName']))

    @staticmethod
    @cache
    def get_env(arch: VsArch):
        if os.environ.get('VSCMD_ARG_HOST_ARCH', None) == arch.name:
            # environment already configured properly
            return None

        installation_info = MSVC.vswhere()
        setup_env = Path(installation_info['installationPath']) / "Common7" / "Tools" / "VsDevCmd.bat"
        cmd = f'cmd.exe /s /c "\"{setup_env}\" -arch={arch.value} >nul 2>&1 && set"'

        result = run(cmd)
        assert result.returncode == 0, "Setting up environment failed"
        return dict([line.split("=", maxsplit=1) for line in result.stdout.splitlines()])

    @classmethod
    @cache
    def discover(cls):
        if platform.system() != "Windows":
            # This compiler is not available on UNIX systems
            return []

        if 'VSCMD_VER' in os.environ:
            logging.debug("Developer VS Shell environment variables found. Skipping discovery.")
            compiler = shutil.which("cl")
            return [] if compiler is None else [Path(compiler)]

        installation_info = MSVC.vswhere()
        if not installation_info:
            # MSVC isn't installed
            return []

        logging.debug("Discovered %s version %s",
                      installation_info['displayName'],
                      installation_info['installationVersion'])

        compilers: list[Path] = []
        for arch in VsArch:
            env = MSVC.get_env(arch)
            compiler = shutil.which("cl", path=env['Path'])
            if compiler is None:
                continue
            compilers.append(Path(compiler))
        return compilers



import logging
from typing import Optional
from pathlib import Path
from tempfile import TemporaryDirectory
from contextlib import nullcontext
import click

from diagtest.log import setup_logger
from diagtest.default import load_compilers, load_languages
from diagtest.driver import Runner
from diagtest.output.terminal import TerminalPrinter
from diagtest.printer import Printer

setup_logger()


@click.command()
@click.option("--list-compilers", type=bool, is_flag=True, default=False, help="Print discovered compilers and exit.")
@click.option("--output", type=Path, default=None, help="Path to build directory")
@click.option("--verbose", type=bool, is_flag=True, default=False, help="Be more verbose")
@click.option("--language", "-l", type=str, default="", help="Target language")
@click.option("--sort-by-assertion", type=bool, default=False, is_flag=True, help="Sort results by assertion, not by compiler")
@click.option("--junit-xml", type=Path, default=None, help="Path to output JUnit style XML test report to")
#@click.option("--json", type=Path, default=None, help="Path to output JSON test report to")
@click.option("--brief", type=bool, default=False, is_flag=True, help="Do not output extra info for failed assertions")
@click.argument("files", type=Path, nargs=-1)
def main(files: list[Path], sort_by_assertion: bool, list_compilers: bool = False, verbose: bool = False, brief: bool = False,
         output: Optional[Path] = None, language: str = "", junit_xml: Optional[Path] = None, json: Optional[Path] = None):
    assert all(file.exists() for file in files), "Please provide a valid file paths"
    if verbose:
        logging.root.setLevel(logging.DEBUG)

    load_languages()
    load_compilers()

    if list_compilers:
        from diagtest.default import dump_compilers
        dump_compilers()
        return

    if not files:
        raise RuntimeError("Input files missing")

    printers: list[Printer] = [TerminalPrinter(brief=brief, assertion_first=sort_by_assertion)]
    if junit_xml:
        from diagtest.output.junit import JUnitPrinter
        printers.append(JUnitPrinter(junit_xml, assertion_first=sort_by_assertion))

    #if json:
    #    from diagtest.output.json import JSONPrinter
    #    printers.append(JSONPrinter(json))

    with nullcontext(output) if output is not None else TemporaryDirectory() as out_path:
        runner = Runner(list(files), Path(out_path), language)
        results = list(runner.run())

    for printer in printers:
        printer.print(results)

    # raise SystemExit(not all(Runner(path, output, language).run() for path in files))



from pathlib import Path

class InterpreterError(Exception):
    ...

class UsageError(InterpreterError):
    def __init__(self, context: tuple[Path, int, int, int], message: str):
        # context[3] is the number of characters processed - we don't care
        self.file, self.line, self.column, *_ = context
        super().__init__(message)
